# Copyright 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from glare.objects import base
from glare.objects.meta import wrappers

Blob = wrappers.BlobField.init


class HeatEnvironment(base.BaseArtifact):

    fields = {
        'environment': Blob(description="Heat Environment text body."),
    }

    @classmethod
    def get_type_name(cls):
        return "heat_environments"
